from apischema.validation.dependencies import (find_dependencies,
                                               find_end_dependencies)


def test_find_dependencies():
    def a_equal_b(self):
        assert self.a == self.b

    assert find_dependencies(a_equal_b) == {"a", "b"}
    assert find_dependencies(int) == set()


def test_find_end_dependencies():
    class Test:
        def __init__(self):
            self.a = 0
            self.b = {}

        def pseudo_validate(self):
            if self.a not in self.method(0):
                yield "error"

        def method(self, arg):
            res = list(self.c)
            if len(res) < arg:
                return self.method(arg - 1)

        @property
        def c(self):
            return self.b.values()

    assert find_end_dependencies(Test, Test.pseudo_validate) == {"a", "b"}



from typing import Dict

from apischema.types import AnyType

_bool_pairs = (
    ("0", "1"),
    ("f", "t"),
    ("n", "y"),
    ("no", "yes"),
    ("false", "true"),
    ("off", "on"),
    ("ko", "ok"),
)
STR_TO_BOOL: Dict[str, bool] = {}
for false, true in _bool_pairs:
    for s, value in ((false, False), (true, True)):
        STR_TO_BOOL[s] = value
        STR_TO_BOOL[s.capitalize()] = value
        STR_TO_BOOL[s.upper()] = value


def coerce(cls: AnyType, data):
    if isinstance(data, cls):
        return data
    if cls is bool and isinstance(data, str):
        return STR_TO_BOOL[data]
    return cls(data)


STR_NONE_VALUES = {""}



from dataclasses import dataclass, field
from typing import Any, Mapping

from apischema import (alias, from_data, get_fields_set, properties, to_data,
                       with_fields_set)


@with_fields_set
@dataclass
class Serialized:
    cls: str = field(metadata=alias("class"))
    to_be_completed: bool = False
    properties: Mapping[str, Any] = field(default_factory=dict,
                                          metadata=properties())


def to_camel_case(s: str) -> str:
    pascal_case = "".join(map(str.capitalize, s.split("_")))
    return pascal_case[0].lower() + pascal_case[1:]


def test_serialized():
    data = {
        "class":  "Foo",
        "field1": 0,
        "field2": ["elt1", "elt2"]
    }
    serialized = from_data(Serialized, data)
    assert serialized == Serialized("Foo", properties={
        "field1": 0,
        "field2": ["elt1", "elt2"]
    })
    assert get_fields_set(serialized) == {"cls", "properties"}  # no to_be_completed
    assert to_data(serialized) == data
    assert to_data(serialized, exclude_unset=False) == {
        "class":           "Foo",
        "to_be_completed": False,
        "field1":          0,
        "field2":          ["elt1", "elt2"]
    }



from dataclasses import (Field as BaseField, InitVar, MISSING, _FIELDS,  # type: ignore
                         dataclass, fields, is_dataclass)
from enum import Enum, auto
from typing import (Any, Callable, Dict, List, Optional, Pattern, Tuple, Type,
                    TypeVar, cast, get_type_hints)

from apischema.alias import ALIAS_METADATA
from apischema.conversion import (INPUT_METADATA, OUTPUT_METADATA, check_converter,
                                  handle_potential_validation, substitute_type_vars)
from apischema.properties import PROPERTIES_METADATA
from apischema.schema import (ANNOTATIONS_METADATA, Annotations, CONSTRAINT_METADATA,
                              Constraint)
from apischema.types import AnyType
from apischema.validation.validator import BaseValidator, VALIDATORS_METADATA, validate

Cls = TypeVar("Cls", bound=Type)


def slotted_dataclass(cls: Cls) -> Cls:
    types = get_type_hints(cls)
    slots = [f.name for f in fields(cls) if not isinstance(types[f.name], InitVar)]
    namespace = cls.__dict__.copy()
    for slot in slots:
        namespace.pop(slot, ...)
    namespace["__slots__"] = slots
    return cast(Cls, type(cls.__name__, cls.__bases__, namespace))


class FieldKind(Enum):
    INIT = auto()
    NORMAL = auto()
    NO_INIT = auto()


@slotted_dataclass
@dataclass
class Field:
    base_field: BaseField
    name: str
    alias: str
    type: AnyType
    input_type: AnyType
    output_type: AnyType
    default: bool
    kind: FieldKind
    annotations: Optional[Annotations]
    constraint: Optional[Constraint]
    input_converter: Optional[Callable[[Any], Any]]
    output_converter: Optional[Callable[[Any], Any]]


FieldCache = Tuple[List[Field], List[Tuple[Pattern, Field]], Optional[Field]]
_input_fields: Dict[Type, FieldCache] = {}
OutputFieldCache = Tuple[List[Field], List[Field]]
_output_fields: Dict[Type, OutputFieldCache] = {}

T = TypeVar("T")


def _add_field_to_lists(obj: T, kind: FieldKind, inputs: List[T], outputs: List[T]):
    if not kind == FieldKind.INIT:
        outputs.append(obj)
    if not kind == FieldKind.NO_INIT:
        inputs.append(obj)


def cache_fields(cls: Type):
    assert is_dataclass(cls)
    types = get_type_hints(cls)
    input_fields: List[Field] = []
    input_patterns: List[Tuple[Pattern, Field]] = []
    input_additional: List[Field] = []
    output_fields: List[Field] = []
    output_patterns: List[Tuple[Pattern, Field]] = []
    output_additional: List[Field] = []
    for field in getattr(cls, _FIELDS).values():
        assert isinstance(field, BaseField)
        type_ = types[field.name]
        if isinstance(type_, InitVar):
            type_ = type_.type  # type: ignore
            kind = FieldKind.INIT
        elif field.init:
            kind = FieldKind.NORMAL
        else:
            kind = FieldKind.NO_INIT
        metadata = field.metadata
        input_type, output_type = type_, type_
        input_converter, output_converter = None, None
        if INPUT_METADATA in metadata:
            param, converter = metadata[INPUT_METADATA]
            param, _ = check_converter(converter, param, ...)  # type: ignore
            param, _ = substitute_type_vars(param, ...)  # type: ignore
            _, converter = handle_potential_validation(..., converter)  # type: ignore
            if VALIDATORS_METADATA in metadata:
                validators: List[BaseValidator] = metadata[VALIDATORS_METADATA]
                conv = converter
                converter = lambda data: validate(conv(data), validators)  # noqa
            input_type, input_converter = param, converter
        if OUTPUT_METADATA in metadata:
            ret, converter = metadata[OUTPUT_METADATA]
            _, ret = check_converter(converter, ..., ret)  # type: ignore
            ret, _ = substitute_type_vars(ret, ...)  # type: ignore
            output_type, output_converter = ret, converter
        new_field = Field(
            base_field=field,
            name=field.name,
            alias=metadata.get(ALIAS_METADATA, field.name),
            type=field.type,
            input_type=input_type,
            output_type=output_type,
            default=(field.default is not MISSING
                     or field.default_factory is not MISSING),  # type: ignore
            kind=kind,
            annotations=metadata.get(ANNOTATIONS_METADATA),
            constraint=metadata.get(CONSTRAINT_METADATA),
            input_converter=input_converter,
            output_converter=output_converter,
        )
        if PROPERTIES_METADATA in metadata:
            pattern = metadata[PROPERTIES_METADATA]
            if not field.init:
                raise TypeError("properties field cannot have init=false")
            if pattern is None:
                if input_additional or output_additional:
                    raise TypeError(f"Multiple additional properties for class {cls}")
                _add_field_to_lists(new_field, new_field.kind,
                                    input_additional, output_additional)
            else:
                _add_field_to_lists((pattern, new_field), new_field.kind,
                                    input_patterns, output_patterns)
        else:
            _add_field_to_lists(new_field, new_field.kind,
                                input_fields, output_fields)
    _input_fields[cls] = (input_fields, input_patterns,
                          input_additional[0] if input_additional else None)
    _output_fields[cls] = (output_fields,
                           [f for p, f in output_patterns] + output_additional)


def get_input_fields(cls: Type) -> FieldCache:
    assert is_dataclass(cls)
    try:
        return _input_fields[cls]
    except KeyError:
        cache_fields(cls)
        return _input_fields[cls]


def get_output_fields_raw(cls: Type) -> OutputFieldCache:
    assert is_dataclass(cls)
    try:
        return _output_fields[cls]
    except KeyError:
        cache_fields(cls)
        return _output_fields[cls]


def get_output_fields(cls: Type) -> FieldCache:
    fields, properties_fields = get_output_fields_raw(cls)
    pattern_fields: List[Tuple[Pattern, Field]] = []
    additional_fields = None
    for field in properties_fields:
        pattern = field.base_field.metadata[PROPERTIES_METADATA]
        if pattern is not None:
            pattern_fields.append((pattern, field))
        else:
            additional_fields = field
    return fields, pattern_fields, additional_fields



"""Kind of typing_extensions for this package"""
__all__ = [
    "Annotated",
    "Literal",
    "NamedTupleMeta",
    "Protocol",
    "TypedDict",
    "_AnnotatedAlias",
    "_GenericAlias",
    "_TypedDictMeta",
    "get_type_hints",
    "set_type_hints",
    "_type_repr"
]

import sys
import types
from dataclasses import is_dataclass
from inspect import isclass
from typing import (Any, Dict, Generic, Mapping, NamedTuple, Optional, Type, TypeVar)


class NoType:
    pass


NO_TYPE = NoType()

if sys.version_info >= (3, 9):
    from typing import Annotated, get_type_hints as gth
else:
    try:
        from typing_extensions import Annotated, get_type_hints as gth
    except ImportError:
        from typing import get_type_hints as gth_


        def gth(obj, globalns=None, localns=None,  # type: ignore # noqa
                include_extras=False):
            return gth_(obj, globalns, localns)


        class MetaAnnotated(type):
            def __getitem__(self, _):
                return NO_TYPE


        class Annotated(MetaAnnotated):  # type: ignore # noqa
            pass
if sys.version_info >= (3, 8):
    from typing import Literal, Protocol, TypedDict
else:
    try:
        from typing_extensions import Literal, Protocol, TypedDict
    except ImportError:
        Literal = NO_TYPE  # type: ignore
        TypedDict = NO_TYPE  # type: ignore
        Protocol = NO_TYPE  # type: ignore

Annotated = Annotated
Literal = Literal
Protocol = Protocol
TypedDict = TypedDict

_T = TypeVar("_T")
_GenericAlias = type(Generic[_T])
_AnnotatedAlias = type(Annotated[_T, ...])
_TypedDictMeta = type(TypedDict)
NamedTupleMeta = type(NamedTuple)

_type_hints: Dict[str, Mapping[str, Type]] = {}

NS = Dict[str, Any]


def get_type_hints(obj, globalns: Optional[NS] = None,
                   localns: Optional[NS] = None):
    try:
        return _type_hints[obj]
    except (KeyError, TypeError):
        types = gth(obj, globalns, localns, include_extras=True)
        try:
            _type_hints[obj] = types
        except TypeError:
            pass
        if isclass(obj) and is_dataclass(obj):
            from apischema.conversion import resolve_fieds_converters
            resolve_fieds_converters(obj)
        return types


set_type_hints = get_type_hints

try:  # pragma: no cover
    from typing import _type_repr  # type:ignore
except ImportError:
    # copy from typing._type_repr in case of ...
    def _type_repr(obj):
        if isinstance(obj, type):
            if obj.__module__ == 'builtins':
                return obj.__qualname__
            return f'{obj.__module__}.{obj.__qualname__}'
        if obj is ...:
            return ('...')
        if isinstance(obj, types.FunctionType):
            return obj.__name__
        return repr(obj)



from typing import Any, Collection, Type

from apischema.typing import _type_repr
from apischema.validation.errors import ErrorMsg


def wrong_type(cls: Type, expected: Type) -> ErrorMsg:
    return f"expected type '{_type_repr(expected)}', got '{_type_repr(cls)}'"


def bad_literal(value: Any, values: Collection[Any]) -> ErrorMsg:
    return f"'{value}' is not one of {values}"



from dataclasses import dataclass
from typing import Any, Generic, TypeVar, Union

from pytest import raises
from typing_extensions import Annotated

from apischema import (Ignored, ValidationError, build_input_schema, from_data, schema,
                       to_data, with_fields_set)


class NoResult:
    pass


NO_RESULT = NoResult()

T = TypeVar("T")


@with_fields_set
@dataclass
class Error(Exception):
    code: int
    description: str
    data: Any = ...


@schema(min_properties=1, max_properties=1)
@with_fields_set
@dataclass
class Result(Generic[T]):
    result: Union[T, Annotated[NoResult, Ignored]] = NO_RESULT
    error: Union[Error, Annotated[None, Ignored]] = None

    def get(self) -> T:
        if self.error is not None:
            raise self.error
        assert not isinstance(self.result, NoResult)
        return self.result


def test_result():
    assert to_data(build_input_schema(Result[int])) == {
        "additionalProperties": False,
        "maxProperties":        1,
        "minProperties":        1,
        "properties":           {
            "result": {"type": "integer"},
            "error": {
                "additionalProperties": False,
                "properties":           {
                    "code":        {"type": "integer"},
                    "description": {"type": "string"},
                    "data":        {}
                },
                "required":             ["code", "description"],
                "type":                 "object"
            },
        },
        "type":                 "object"
    }

    data = {"result": 0}
    with raises(ValidationError):
        from_data(Result[str], data)
    result = from_data(Result[int], data)
    assert result.get() == 0
    assert to_data(result) == {"result": 0}

    error = from_data(Result, {"error": {"code": 42, "description": "..."}})
    with raises(Error) as err:
        error.get()
    assert err.value == Error(42, "...")



from dataclasses import dataclass, field, fields

from pytest import raises

from apischema import (get_fields_set, mark_set_fields, unmark_set_fields,
                       with_fields_set)
from apischema.fields import FIELDS_SET_ATTR, NoDefault, get_default


@with_fields_set
@dataclass
class Data:
    without_default: int
    with_default: int = 0
    with_default_factory: int = field(default_factory=lambda: 0)


@dataclass
class Inherited(Data):
    other: int = 42


@with_fields_set
@dataclass
class DecoratedInherited(Data):
    other: int = 42


def test_default():
    without_default, with_default, with_default_factory = fields(Data)
    with raises(NoDefault):
        get_default(without_default)
    assert get_default(with_default) == 0
    assert get_default(with_default_factory) == 0


def test_fields_set():
    with raises(ValueError):
        get_fields_set(object())

    assert get_fields_set(Data(0)) == {"without_default"}
    assert get_fields_set(Data(without_default=0)) == {"without_default"}
    assert get_fields_set(Data(0, 1)) == {"without_default", "with_default"}

    data = Data(0)
    data.with_default = 1
    assert get_fields_set(data) == {"without_default", "with_default"}
    unmark_set_fields(data, "without_default")
    assert get_fields_set(data) == {"with_default"}
    mark_set_fields(data, "with_default_factory")
    assert get_fields_set(data) == {"with_default", "with_default_factory"}
    mark_set_fields(data, "with_default", overwrite=True)
    assert get_fields_set(data) == {"with_default"}
    data.__dict__.pop(FIELDS_SET_ATTR)
    assert get_fields_set(data) == {"without_default", "with_default",
                                    "with_default_factory"}
    unmark_set_fields(data, "without_default")
    assert get_fields_set(data) == {"with_default", "with_default_factory"}
    data.__dict__.pop(FIELDS_SET_ATTR)
    mark_set_fields(data, "with_default")
    assert get_fields_set(data) == {"without_default", "with_default",
                                    "with_default_factory"}
    with raises(ValueError):
        mark_set_fields(data, "not_a_field")

    assert get_fields_set(DecoratedInherited(0, other=0)) == {"without_default",
                                                              "other"}
    assert get_fields_set(Inherited(0, other=0)) == {
        "without_default", "with_default", "with_default_factory", "other"
    }



from typing import Generic, TypeVar

from apischema.conversion import substitute_type_vars

T = TypeVar("T")
U = TypeVar("U")
V = TypeVar("V")


class A(Generic[T, U]):
    pass


class B(Generic[T]):
    pass


def test_substitute_type_vars():
    assert substitute_type_vars(A[int, V], B[V]) == (A[int, U], B[U])



from dataclasses import Field
from functools import wraps
from inspect import Parameter, isfunction, isgeneratorfunction, signature
from types import MethodType
from typing import (AbstractSet, Any, Callable, Iterable, Optional, Sequence, TypeVar,
                    cast, overload)

from apischema.types import AnyType, DictWithUnion, Metadata
from apischema.typing import Protocol
from apischema.utils import PREFIX
from apischema.validation.dependencies import find_end_dependencies
from apischema.validation.errors import (ValidationError, build_from_errors, exception,
                                         merge)
from apischema.validation.mock import NonTrivialDependency

# use attribute instead of global dict in order to be inherited
VALIDATORS_ATTR = f"{PREFIX}validators"


class ValidatorFunc(Protocol):
    @overload
    def __call__(self, _obj):
        ...

    @overload
    def __call__(self, _obj, **kwargs):
        ...

    def __call__(self, _obj, **kwargs):
        ...


class BaseValidator(Protocol):
    validate: ValidatorFunc


class SimpleValidator:
    def __init__(self, func: ValidatorFunc):
        self.func = func
        parameters = signature(func).parameters
        validate = func
        if not any(param.kind == Parameter.VAR_KEYWORD
                   for param in parameters.values()):
            if not parameters:
                raise TypeError("validator must have at least one parameter")
            params = set(list(parameters)[1:])
            wrapped = validate

            def validate(_obj, **kwargs):
                kwargs2 = {k: v for k, v in kwargs.items() if k in params}
                return wrapped(_obj, **kwargs2)

        if isgeneratorfunction(func):
            wrapped = validate

            def validate(_obj, **kwargs):
                errors = [*validate(_obj, **kwargs)]
                if errors:
                    raise build_from_errors(errors)

        self.validate = cast(ValidatorFunc, validate)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def can_be_called(self, fields: AbstractSet[str]) -> bool:
        return False


class Validator(SimpleValidator):
    def __init__(self, func: Callable):
        if isgeneratorfunction(func):
            @wraps(func)
            def validate(_obj, **kwargs):
                errors = []
                try:
                    errors.extend(func(_obj, **kwargs))
                except ValidationError:
                    raise RuntimeError("Validation error raised in generator validator")
                except Discard as err:
                    err.error = build_from_errors(errors)
                    raise
                if errors:
                    raise build_from_errors(errors)

            validate.__annotations__.pop("return", ...)
            super().__init__(validate)
        else:
            super().__init__(func)
        wraps(func)(self)
        self.func = func

    def __get__(self, instance, owner):
        return self if instance is None else MethodType(self.func, instance)

    def __set_name__(self, owner, name: str):
        self.owner = owner
        self.dependencies = find_end_dependencies(owner, self.func)
        setattr(owner, VALIDATORS_ATTR, (*get_validators(owner), self))

    def can_be_called(self, fields: AbstractSet[str]) -> bool:
        return all(dep in fields for dep in self.dependencies)


Func = TypeVar("Func", bound=Callable)


def get_validators(obj) -> Sequence[Validator]:
    return getattr(obj, VALIDATORS_ATTR, ())


def add_validator(cls: AnyType, *validators: ValidatorFunc):
    for func in validators:
        validator = SimpleValidator(func)
        setattr(cls, VALIDATORS_ATTR, (*get_validators(cls), validator))


class Discard(Exception):
    def __init__(self, *fields):
        if not all(isinstance(f, Field) for f in fields):
            raise TypeError("Only fields can be discarded")
        self.fields: AbstractSet[str] = {f.name for f in fields}
        self.error: Optional[ValidationError] = None


T = TypeVar("T")


def validate(_obj: T, _validators: Sequence[BaseValidator] = None, **kwargs) -> T:
    if _validators is None:
        _validators = get_validators(_obj)
    if not _validators:
        return _obj
    error: Optional[ValidationError] = None
    while True:
        for i, validator in enumerate(_validators):
            try:
                validator.validate(_obj, **kwargs)
            except ValidationError as err:
                error = merge(error, err)
            except Discard as err:
                if err.error is None:
                    raise RuntimeError("Discard can only be raised in class Validator")
                error = merge(error, err.error)
                _validators = [v for v in _validators[i + 1:]
                               if not isinstance(v, Validator)
                               or not v.dependencies & err.fields]
                break
            except NonTrivialDependency as exc:
                assert isinstance(validator, Validator)
                exc.validator = validator
                raise
            except Exception as err:
                error = merge(error, ValidationError([exception(err)]))
        else:
            break
    if error is not None:
        raise error
    return _obj


V = TypeVar("V", bound=Callable)


@overload
def validator(func: V) -> V:
    ...


@overload
def validator(field: Any, *, discard: bool = True) -> Callable[[V], V]:
    ...


def validator(arg=None, field=None, *, discard=True):
    if isfunction(arg):
        return Validator(arg)
    if arg is None and field is None or arg is not None and field is not None:
        raise ValueError("Bad use of validator")
    field = field or arg
    if not isinstance(field, Field):
        raise TypeError("validator argument must be a field")

    def decorator(func):
        @wraps(func)
        def wrapper(self, **kwargs):
            has_error = False
            for error in func(self, **kwargs):
                has_error = True
                if isinstance(error, str):
                    yield field, error
                else:
                    path, msg = error
                    if isinstance(path, Iterable) and not isinstance(path, str):
                        yield (field, *path), msg
                    else:
                        yield (field, path), msg
            if has_error and discard:
                raise Discard(field)

        return validator(wrapper)

    return decorator


VALIDATORS_METADATA = f"{PREFIX}validators"


def field_validator(*validators: ValidatorFunc) -> Metadata:
    return DictWithUnion({VALIDATORS_METADATA: list(map(SimpleValidator, validators))})
